# import cakecms api module even if it is not installed
import sys
import os

sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))

import cakecms
import argparse

'''
USAGE: Seat on testing #3, order by #5-points descending
python change_seating.py "http://localhost/cakecms" prog1 2q4q3dsLrnFsXf8AMFSKEZj3 3 --order 5
'''


def order_students_by_points(cms, students, testing_id, desc=True):
	results = cms.testingresults_index_by_student_id(testing_id=testing_id)
	students.sort(key=lambda studentid: results[studentid]['points'])
	if desc: students = students[::-1]
	return students


def order_students_by_matriculation(cms, students, desc=True):
	matriculations = cms.students_index()['students']
	matriculations = {student['Student']['id']: int(student['Student']['matriculation']) for student in matriculations}
	students.sort(key=lambda studentid: matriculations[studentid])
	if desc: students = students[::-1]
	return students


def seat_bad_students_before_good_students(students, seats):
	"""
	:param students: (Ordered) student ids
	:param seats: (Ordered) seats: By room, then front to back, then left to right
	:return: list of new seatings: {student_id: ?, room_id: ?, row: ?, seat: ?, rownumber: ?, seatnumber: ?} (optional: code)
	"""
	seat_set = set(seats)
	front_seats = [(room, row, seat, rownum, seatnum) for room, row, seat, rownum, seatnum in seats
				   if not (room, row - 1, seat, rownum - 1, seatnum) in seat_set]
	back_seats = [(room, row, seat, rownum, seatnum) for room, row, seat, rownum, seatnum in seats
				  if (room, row - 1, seat, rownum - 1, seatnum) in seat_set]
	front_students = students[len(back_seats):]
	back_students = students[:len(back_seats)]
	# bug-hunting
	assert len(front_seats) != 0 and len(front_seats) != len(seats)
	assert len(back_seats) != 0 and len(back_seats) != len(seats)
	assert len(seats) == len(seat_set)
	assert len(front_seats) == len(front_students)
	assert len(back_seats) == len(back_students)
	assert len(front_seats) + len(back_seats) == len(seats)
	assert len(seats) == len(students)
	# bug-hunting end
	return [{'student_id': student, 'room_id': room, 'row': row, 'seat': seat, 'rownumber': rownum, 'seatnumber': seatnum}
			for (room, row, seat, rownum, seatnum), student in zip(front_seats + back_seats, front_students + back_students)]


def assign_seating(cms, testing_id, student_order_function, seating_function):
	"""
	:param cms:
	:param testing_id:
	:param list[int] => list[int] student_order_function:
	:param (list[int], list[(int, int, int)]) => list[dict] seating_function:
	:return:
	"""
	# retrieve students and seats
	admissions = cms.admissions_index(testing_id=testing_id)
	print 'Retrieved', len(admissions), 'seated students'

	# split in students / seats
	students = [adm['Admission']['student_id'] for adm in admissions]
	seats = [(adm['Admission']['room_id'], adm['Admission']['row'], adm['Admission']['seat'], adm['Admission']['rownumber'],
			  adm['Admission']['seatnumber'])
			 for adm in admissions]
	assert len(seats) == len(set(seats))

	# order seats
	# TODO if you want a special room order, apply it HERE
	seats.sort()

	# order students
	students = student_order_function(students)

	# assign each student a seat
	data = seating_function(students, seats)

	# save
	'''
	admission_ids = {adm['Admission']['student_id'] : adm['Admission']['id'] for adm in admissions}
	print 'Saving seating. This must not be interrupted.'
	print 'If saving is interrupted, you have to run the CMS distribution before running this script again.'
	for i, entry in enumerate(data):
		id = admission_ids[entry['student_id']]
		entry['testing_id'] = testing_id
		cms.assert_success(cms.admissions_edit(id, entry))
		print 'Saving {:6.1%} completed...  ({}/{})'.format((i+1.0)/len(data), i+1, len(data))
	'''
	for element in data:
		element['testing_id'] = testing_id
	cms.assert_success(cms.admissions_edit_many(data))
	print 'Done!'


def on_localhost():
	cms = cakecms.CakeCMS('http://localhost/cakecms', '2q4q3dsLrnFsXf8AMFSKEZj3', 'prog1')
	cms.debug = True
	assign_seating(cms, 19,
				   lambda students: order_students_by_points(cms, students, 2, True),  # how to order students
				   seat_bad_students_before_good_students  # how to seat students
				   )


def main():
	# Run <script> -h to list parameters
	parser = argparse.ArgumentParser(description='Change seating order for an exam.')
	parser.add_argument('url', help='CakeCMS base url')
	parser.add_argument('course', help='CakeCMS course shortname (e.g. "prog1_16")')
	parser.add_argument('token', help='CakeCMS API token')
	parser.add_argument('id', type=int, help='Id of the testing whose seating you want to adjust')
	parser.add_argument('--order', action='store', default='matriculation', metavar='TESTING-ID', help='Order students by testing points')
	args = parser.parse_args()
	cms = cakecms.CakeCMS(args.url, args.token, args.course)
	cms.debug = True

	# If you want your own order function, add it here
	if args.order == 'matriculation':
		student_order_function = lambda students: order_students_by_matriculation(cms, students)
	else:
		student_order_function = lambda students: order_students_by_points(cms, students, int(args.order))
	# If you want your own seating distribution, add it here
	seating_function = seat_bad_students_before_good_students

	assign_seating(cms, args.id, student_order_function, seating_function)


if __name__ == '__main__':
	main()
