from cakecms import CakeCMS
import json
import unicodecsv as csv
import hmac
import hashlib
from base64 import *

csv.register_dialect('cms', delimiter=',', lineterminator='\n')
csv.register_dialect('realexcel', delimiter=';', lineterminator='\r\n', quoting=csv.QUOTE_ALL)
BOM = '\xEF\xBB\xBF'

class StudentTable:
	def __init__(self, content):
		self.text = content
		self.dialect = csv.get_dialect('cms' if content.count(',') >= content.count(';') else 'realexcel')
		reader = csv.reader(content.splitlines(), dialect=self.dialect)
		self.columns = reader.next()
		self.columns_lc = [col.lower() for col in self.columns]
		self.anonymized = 'first name' not in self.columns_lc
		self.content = [row for row in reader]

	def __str__(self):
		return 'StudentTable {{\n\tcolumns: {}\n\tcontent: {} lines\n\tanonym: {}\n}}'.format(self.columns, len(self.content), self.anonymized)

	def __repr__(self):
		return str(self)

	def anonymize(self, hmac_key = '', hmac_hash=hashlib.md5):
		if self.anonymized:
			return
		def anonymize_matriculation(msg):
			"""
			Reference: SELECT ENCODE(HMAC(matriculation::TEXT,'key','md5'),'hex');
			"""
			msg = str(int(msg))  # remove trailing zeros
			return hmac.new(hmac_key, msg, hmac_hash).hexdigest()
		# remove name columns
		self.columns = self.columns[0:1]+self.columns[3:]
		self.columns_lc = self.columns_lc[0:1]+self.columns_lc[3:]
		self.content = [[anonymize_matriculation(row[0])] + row[3:] for row in self.content]
		self.anonymized = True

	def save(self, fname):
		with open(fname, 'wb') as f:
			writer = csv.writer(f, dialect=self.dialect)
			writer.writerow(self.columns)
			for row in self.content:
				writer.writerow(row)

	def save_excel(self, fname):
		with open(fname, 'wb') as f:
			f.write(BOM)
			writer = csv.writer(f, dialect='realexcel')
			writer.writerow(self.columns)
			for row in self.content:
				writer.writerow(row)

	def fill_missing_matriculations(self, reference):
		def get_keys(row, matr_idx, name1_idx, name2_idx, gender_idx, term1_idx, term2_idx):
			keys = []
			# optional - remove second name
			fname = row[name1_idx].split(' ')
			for i in range(0, len(fname)):
				name1 = ' '.join(fname[0:i+1])
				key = (name1+','+row[name2_idx]+','+row[gender_idx]+',{},{}').lower()
				for j in (0, 1):
					for k in (0,1):
						keys.append(key.format(int(row[term1_idx])+j, int(row[term2_idx])+k))
			return keys

		# build reference mapping: "key" -> matriculation
		mapping = {}
		ref_matr_idx   = reference.columns_lc.index('matriculation')
		ref_name1_idx  = reference.columns_lc.index('first name')
		ref_name2_idx  = reference.columns_lc.index('last name')
		ref_gender_idx = reference.columns_lc.index('gender')
		ref_term1_idx  = reference.columns_lc.index('term of subject')
		ref_term2_idx  = reference.columns_lc.index('term of study')
		for row in reference.content:
			matr = row[ref_matr_idx]
			for key in get_keys(row, ref_matr_idx, ref_name1_idx, ref_name2_idx, ref_gender_idx, ref_term1_idx, ref_term2_idx):
				mapping[key] = matr

		# map all unknown students
		corrected_students = 0
		unknown_students = []
		matr_idx   = self.columns_lc.index('matriculation')
		name1_idx  = self.columns_lc.index('first name')
		name2_idx  = self.columns_lc.index('last name')
		gender_idx = self.columns_lc.index('gender')
		term1_idx  = self.columns_lc.index('term of subject')
		term2_idx  = self.columns_lc.index('term of study')
		for row in self.content:
			if row[matr_idx].startswith('9'):
				keys = get_keys(row, matr_idx, name1_idx, name2_idx, gender_idx, term1_idx, term2_idx)
				known_keys = [key for key in keys if key in mapping]
				if len(known_keys) > 0:
					row[matr_idx] = mapping[known_keys[0]]
					corrected_students += 1
				else:
					unknown_students.append((row[matr_idx], keys, row))
		return (corrected_students, unknown_students)

	def filter_by_matriculations(self, matriculations):
		assert not self.anonymized
		matr_idx = self.columns_lc.index('matriculation')
		self.content = [row for row in self.content if row[matr_idx] in matriculations]

	def filter_out_matriculations(self, matriculations):
		assert not self.anonymized
		matr_idx = self.columns_lc.index('matriculation')
		self.content = [row for row in self.content if row[matr_idx] not in matriculations]

	def filter_invalid_rows(self):
		# Filter rows that have an invalid matriculation
		assert not self.anonymized
		matr_idx = self.columns_lc.index('matriculation')
		self.content = [row for row in self.content if not row[matr_idx].startswith('9')]

	@staticmethod
	def from_file(fname):
		with open(fname, 'rb') as f:
			return StudentTable(f.read())

	@staticmethod
	def from_cms(cms, course=None):
		columns = cms.students_index_get_cols(course=course)
		anonymous_columns = [k for k, col in columns.items() if
							 col != 'Student.id' and
							 col != 'Student.matriculation' and (
								 col.startswith('Student.') or
								 col.startswith('Statistic.') or
								 col.startswith('Testingresult')
							 )]
		# matriculation must be first column
		columns = ['Sm', 'Uf', 'Ul'] + anonymous_columns
		# retrieve csv
		csv = cms.students_index_csv(course=course, columns=columns)
		return StudentTable(csv)



# ===== TEST =====

if __name__ == '__main__':
	with open('tokens.json', 'rb') as f:
		TOKENS = json.loads(f.read())

	# TEST 1 - import
	def import_tables():
		# Read from VorkursCMS
		vorkurscms = CakeCMS('https://vorkurs.cs.uni-saarland.de/cms/', TOKENS['vorkurs'])
		vorkurscms.debug = True
		table = StudentTable.from_cms(vorkurscms, 'ss16')
		table.save('vorkurs.csv')
		with open('vorkurs-orig.csv', 'wb') as f:
			f.write(table.text)
		# Read from Prog1
		dcms = CakeCMS('https://dcms.cs.uni-saarland.de/', TOKENS['dcms'])
		dcms.debug = True
		table = StudentTable.from_cms(dcms, 'prog1_16')
		table.save('prog1.csv')
		with open('prog1-orig.csv', 'wb') as f:
			f.write(table.text)

	#import_tables()
	vorkurs_table = StudentTable.from_file('vorkurs.csv')
	prog1_table = StudentTable.from_file('prog1.csv')
	#vorkurs_table.save_excel('vorkurs-excel.csv')
	#prog1_table.save_excel('prog1-excel.csv')


	# TEST 2 - join
	# Join Vorkurs with Prog1
	(corrected, unknown) = vorkurs_table.fill_missing_matriculations(prog1_table)
	print 'Corrected', corrected, 'student matriculations,', len(unknown), 'not found in refernce,', len(vorkurs_table.content),'total'
	for matr, info, row in unknown:
		print '-', matr, info

	print '\n'.join([str((matr, info, row)) for (matr, info, row) in unknown if row[18]])
	vorkurs_table.save_excel('vorkurs-joined-excel.csv')

	# TEST 3.1 - filter Matthias Maurer
	prog1_table.filter_by_matriculations(['0042420'])

	# TEST 3.2 - anonymize both
	vorkurs_table.anonymize('abcd')
	prog1_table.anonymize('abcd')


	# TEST 4 - save
	vorkurs_table.save('anonymized-vorkurs.csv')
	vorkurs_table.save_excel('anonymized-vorkurs-excel.csv')
	prog1_table.save('anonymized-prog1.csv')
	prog1_table.save_excel('anonymized-prog1-excel.csv')
