from __future__ import unicode_literals
import sys
import os
from os.path import join
import errno
import shutil
import yaml
import json
from collections import MutableMapping

from cakecms import CakeCMS
from cms_student_tables import StudentTable

import logging
logging.basicConfig(filename='error.log', level=logging.WARN)


"""
- Extracts all student data ("tables") from various cms systems
- Restores missing matriculation numbers
- Anonymizes the tables
- Saves all information as standard utf-8 csv files (and excel-compatible ones)

Configuration must be yaml or json. Refer to example.yaml for the format.
Files must be called 'default.yaml' (public configuration) and 'tokens.yaml' (the secret values).

USAGE: python extract-students.py [config-files...] [-i]
PARAMETERS:
-i   Read configuration (json or yaml) from stdin
"""


def warn(msg):
	print '[WARN]', msg

error_count = 0
def error(msg, e):
	global error_count
	error_count += 1
	print '[ERROR] {}: {}'.format(msg, e.message)
	logging.exception(msg)



def read_config(defaults = []):
	# Read all given files (if existing) and everything passed by command line
	# -i reads additional config from stdin
	config = {}

	def rec_merge(d1, d2):
		for k, v in d1.items():
			if k in d2:
				if all(isinstance(e, MutableMapping) for e in (v, d2[k])):
					rec_merge(v, d2[k])
					d2[k] = v
		d1.update(d2)

	def add_config(content):
		if content.strip().startswith('{'):
			d = json.loads(content)
		else:
			d = yaml.safe_load(content)
		rec_merge(config, d)

	for arg in defaults + sys.argv[1:]:
		if os.path.exists(arg) and os.path.isfile(arg):
			# read file
			with open(arg, 'rb') as f:
				add_config(f.read())
		elif arg == '-i':
			add_config(sys.stdin.read())
	return config


def assert_config(config):
	"""
	Checks if the given configuration is valid
	:param config:
	"""
	if len(config) == 0:
		raise Exception('No configuration files found!')
	for system in config['systems']:
		assert system in config['tokens']
	valid = set()
	for system, courses in config['courses'].items():
		assert system in config['systems']
		for course in courses:
			valid.add(course+'@'+system)
	if 'fill-missing-matriculations' not in config or not config['fill-missing-matriculations']:
		config['fill-missing-matriculations'] = {}
	for k, v in config['fill-missing-matriculations'].items():
		assert k in valid
		assert v in valid




def prepare(config):
	"""
	Prepares the CMS connection instances, and creates the output folders
	:param config:
	:return: a dictionary <cms-name> => <cms-connection>
	"""
	def mkdir_p(path):
		try:
			os.makedirs(path)
		except OSError as exc:  # Python >2.5
			if exc.errno != errno.EEXIST or not os.path.isdir(path):
				raise

	# create output folders
	for k, v in config['output'].items():
		if not os.path.exists(v):
			mkdir_p(v)
	# create cms connections
	cms_connections = {}
	for system, url in config['systems'].items():
		cms = CakeCMS(url, config['tokens'][system])
		cms.debug = True
		cms_connections[system] = cms
	return cms_connections



def load_all_courses(config, cms_connections):
	"""
	Connect to all cms instances, and download course information (.json) and student data (.csv) from each course.
	Files are saved to "default" output path and cached there. If you want to download again, delete the old files first.
	:param config:
	:param cms_connections:
	:return: dictionary containing all tables (from cms or from file) as StudentTable instance. <systemname> => <table>
	"""
	print '--- Loading tables from {} servers ---'.format(len(config['courses']))
	out = config['output']['default']
	tables = {}
	for system, courses in config['courses'].items():
		for course in courses:
			# Check if files already exist
			ident = course+'@'+system
			if os.path.isfile(join(out, ident+'.calendar.json')) and os.path.isfile(join(out, ident+'.csv')):
				if os.path.getsize(join(out, ident+'.csv')) > 0:
					print 'Found:',ident
					tables[ident] = StudentTable.from_file(join(out, ident+'.csv'))
					continue
			try:
				print 'Loading:', ident
				# Download course info
				coursedata = cms_connections[system].courses_get(shortname=course)
				if coursedata:
					with open(join(out, ident+'.course.json'), 'wb') as f:
						f.write(json.dumps(coursedata, indent=4))
				else:
					warn('Could not get course info for '+ident)

				# Download calendar
				calendar_data = cms_connections[system].calendar_events_index(course=course)
				if calendar_data:
					with open(join(out, ident+'.calendar.json'), 'wb') as f:
						f.write(json.dumps(calendar_data, indent=4))
				else:
					warn('Could not get calendar for '+ident)

				# Download student table
				tables[ident] = StudentTable.from_cms(cms_connections[system], course)
				tables[ident].save(join(out, ident+'.csv'))

			except Exception, e:
				error('Could not retrieve '+ident, e);

	print len(tables),'tables read.\n\n'
	return tables


def table_linking(config, tables):
	"""
	Restores missing matriculations by linking multiple courses together.
	A missing matriculation is filled out if a student with the same name, gender and similar semester count is found.
	Tables are modified in-place, but not saved.
	:param config:
	:param tables:
	"""
	if len(config['fill-missing-matriculations']) == 0:
		return
	out = config['output']['default']
	for ident, fixes in config['fill-missing-matriculations'].items():
		try:
			if isinstance(fixes, basestring):
				fixes = [fixes]
			print '--- Trying to restore matriculations of {} ---'.format(ident)
			corrected = 0
			for fix in fixes:
				corrected_students, unknown_students = tables[ident].fill_missing_matriculations(tables[fix])
				corrected += corrected_students
			print corrected, 'matriculations have been corrected,', len(unknown_students),'students could not be matched.'
			if len(unknown_students) > 0:
				print 'You can try to give these students their matriculation manually (edit', join(out, ident+'.csv'), 'and re-run)'
				for mtknr, _, row in unknown_students:
					try:
						print '- '+mtknr+' '+' '.join(row[1:3]).encode()
					except UnicodeEncodeError, e:
						print '-',mtknr, u' '.join(row[1:3]).encode('ascii', 'replace')

		except Exception, e:
			error('Can\'t fix matriculations for '+ident, e)
		print '\n\n'



def table_filtering(config, tables):
	for key, table in tables.items():
		table.filter_invalid_rows()
	if 'filter' in config and 'matriculation'in config['filter']:
		print '--- Filtering ---'
		for key, table in tables.items():
			table.filter_by_matriculations(config['filter']['matriculation'])
			print 'Filtered', key, ':', len(table.content), 'students remaining'
		print '\n\n'



def table_anonymization(config, tables):
	"""
	Anonymizes all tables (in-place), if enabled in the settings
	:param config:
	:param tables:
	"""
	if config['anonymization']['enabled'] == 1 or config['anonymization']['enabled'] == 'true':
		print '--- Anonymizing {} tables ---'.format(len(tables))
		for key, table in tables.items():
			table.anonymize(hmac_key=str(config['anonymization']['secret']))
		print 'All tables anonymized.\n\n'


def save_anonymized(config, tables):
	"""
	Save the (previously anonymized) tables to the 'anonymized' output directory, including all json course information.
	:param config:
	:param tables:
	"""
	print '--- Exporting {} tables ---'.format(len(tables))
	datapath = config['output']['default']
	out = config['output']['anonymized']
	for ident, table in tables.items():
		print 'Saving',ident,'... ',
		# save csv's
		table.save(join(out, ident+'.csv'))
		table.save_excel(join(out, ident+'.excel.csv'))
		print '[OK]'
	# save json's
	print 'Copying structural information... ',
	for fname in os.listdir(datapath):
		if fname.endswith('.json'):
			shutil.copy(join(datapath, fname), out)
	print '[OK]'
	if error_count == 0:
		print '\n[ DONE ]   You can now pack and send "{}"\n\n'.format(out)
	else:
		print '\n[ DONE ]   Errors:', error_count





def main():
	try:
		# read configuration
		CONFIG = read_config(['default.yaml', 'default.json', 'tokens.yaml', 'tokens.json'])
		assert_config(CONFIG)
		# 0. Prepare everything
		cms_connections = prepare(CONFIG)
		# 1. Load all courses
		tables = load_all_courses(CONFIG, cms_connections)
		# 2. Table processing
		table_linking(CONFIG, tables)
		table_filtering(CONFIG, tables)
		# 3. Table anonymization
		table_anonymization(CONFIG, tables)
		# 4. Save anonymized tables
		save_anonymized(CONFIG, tables)
		# keep window open (if bundled exe)
		if getattr(sys, 'frozen', False):
			raw_input("Press Enter to terminate...")
	except Exception, e:
		error('Error', e)
		raise




if __name__ == '__main__':
	main()
